#!/ usr/bin/env python
# coding: utf-8
# 
# This code generates a CNF formula for a Graph Coloring problem

import sys
import networkx as nx
import matplotlib.pyplot as plt
from pysat.solvers import Glucose3

def exactly_one(vars):
    """
    Return clauses that enforce exactly one variable is True.

    Args:
        vars (list): list of variables for which we only want one to be True

    Returns:
        clauses: list of clauses (in CNF format).
    """
    clauses = []

    # At least one is True
    clauses.append(vars)

    # At most one is True (pairwise negative clauses)
    for i in range(len(vars)):
        for j in range(i + 1, len(vars)):
            clauses.append([-vars[i], -vars[j]])

    return clauses

def varnum(node, num_colors, color):
    """Map the abstract variable to a single variable number."""
    return node * num_colors + color


def graph_coloring_cnf(graph, num_colors):
    """
    Convert a graph coloring problem into a CNF formula in DIMACS format.
    The graph is represented as an adjacency list (dictionary), where keys are nodes
    and values are lists of adjacent nodes.

    Args:
        graph: dict, adjacency list representation of the graph.
        num_colors: int, number of colors available.

    Returns:
        clauses: list of clauses in CNF format.
    """
    clauses = []

    # Each node must have exactly one color
    for node in graph:
        clauses.extend(exactly_one([varnum(node, num_colors, color) for color in range(1, num_colors + 1)]))

    # Adjacent nodes must not share the same color
    for node in graph:
        for neighbor in graph[node]:
            if node < neighbor:  # Avoid duplicate constraints
                for color in range(1, num_colors + 1):
                    clauses.append([-varnum(node, num_colors, color), -varnum(neighbor, num_colors, color)])

    return clauses

def write_dimacs(clauses, num_vars, filename):
    """
    Write the set of clauses to a DIMACS file.

    Args:
        clauses (list): clauses describing the CNF formula 
        num_vars (int): number of variables
        filename (string): filename 
    """
    with open(filename, 'w') as f:
        f.write(f"p cnf {num_vars} {len(clauses)}\n")
        for clause in clauses:
            f.write(" ".join(map(str, clause)) + " 0\n")

def solve_graph_coloring(graph, num_colors):
    """
    Solve the graph coloring problem using a SAT solver.

    The graph is assumed to be given as dictionnary encoding an adjacency list where the nodes are numbered
    consecutively from 0 or 1. For instance:

    graph = {
        0: [1, 2],
        1: [0, 2, 3],
        2: [0, 1, 3, 4],
        3: [1, 2, 4],
        4: [2, 3]
    } 

    Args:
        graph: dict, adjacency list representation of the graph.
        num_colors: int, number of colors available.

    Returns:
        solution: dict, mapping of nodes to their assigned colors, or None if no solution exists.
    """
    clauses = graph_coloring_cnf(graph, num_colors)

    solver = Glucose3()
    for clause in clauses:
        solver.add_clause(clause)

    if solver.solve():
        model = solver.get_model()
        solution = {}
        for var in model:
            if var > 0:  # Only consider positive literals
                node = (var - 1) // num_colors
                color = (var - 1) % num_colors + 1
                solution[node] = color
        return solution
    else:
        print("No solution exists.")
        return None

def print_solution(solution):
    """
    Display the graph coloring solution in a readable format.

    Args:
        solution: dict, mapping of nodes to their assigned colors.
    """
    if not solution:
        return

    print("Graph Coloring Solution:")
    for node, color in sorted(solution.items()):
        print(f"Node {node}: Color {color}")

def display_solution(solution, graph):
    """
    Display the graph coloring solution visually.

    Args:
        solution: dict, mapping of nodes to their assigned colors.
        graph: dict, adjacency list representation of the graph.
    """
    if not solution:
        return

    # Create a graph using NetworkX
    G = nx.Graph()
    for node, neighbors in graph.items():
        for neighbor in neighbors:
            G.add_edge(node, neighbor)

    # Assign colors to nodes based on the solution
    color_map = [solution[node] for node in G.nodes]

    # Draw the graph
    plt.figure(figsize=(8, 6))
    nx.draw(
        G,
        with_labels=True,
        node_color=color_map,
        node_size=500,
        cmap=plt.cm.rainbow,
        font_weight='bold'
    )
    plt.title("Graph Coloring Solution")
    plt.show()
    

def read_graph_from_dimacs(filename):
    """
    Read a graph from a DIMACS file.

    Args:
        filename: str, path to the DIMACS file.

    Returns:
        graph: dict, adjacency list representation of the graph.
        num_colors: int, number of colors available (if specified, otherwise None).
    """
    graph = {}

    with open(filename, 'r') as f:
        for line in f:
            line = line.strip()
            if line.startswith('c'):  # Comment line, skip it
                continue
            elif line.startswith('p'):  # Problem line
                parts = line.split()
                num_nodes = int(parts[2])
                for i in range(num_nodes):
                    graph[i] = []
            elif line.startswith('e'):  # Edge line
                parts = line.split()
                node1 = int(parts[1]) - 1  # Convert to 0-based index
                node2 = int(parts[2]) - 1  # Convert to 0-based index
                graph[node1].append(node2)
                graph[node2].append(node1)

    return graph

def write_graph_to_dimacs(graph, filename):
    """
    Write a graph to a DIMACS file.

    Args:
        graph: dict, adjacency list representation of the graph.
        filename: str, path to the DIMACS file.
    """
    with open(filename, 'w') as f:
        f.write(f"p edge {len(graph)} {sum(len(neighbors) for neighbors in graph.values()) // 2}\n")
        for node, neighbors in graph.items():
            for neighbor in neighbors:
                if node < neighbor:  # Avoid duplicate edges
                    f.write(f"e {node + 1} {neighbor + 1}\n")


if __name__ == "__main__":

    if len(sys.argv) != 2:
        print("Usage: python3 graph-coloring.py <input_file>")
        sys.exit(1)

    input_file = sys.argv[1]

    # Example graph with 5 nodes
    example_graph = read_graph_from_dimacs(input_file)

    num_colors = 3  # Number of colors to use

    # Solve the graph coloring problem
    solution = solve_graph_coloring(example_graph, num_colors)

    # Display the solution
    display_solution(solution, example_graph)